//: Playground - noun: a place where people can play

import Cocoa

var str = "Hello, playground"

enum Happiness1 {                       // our enum with associated values
    case Sad
    case Happy(ofGrade: Int)
}

let mood1 = Happiness1.Happy(ofGrade: 3)    // lets create an instance
// let data1 = NSKeyedArchiver.archivedData(withRootObject: mood1)  // that doesn't work: runtime error

// ---------------------------------------------------------------------

// MARK: - Property list conversion protocol

typealias PropertyList = [String : Any]

protocol PropertyListReadable {         // this protocol can be applied to any struct or enum
    func propertyListRepresentation() -> PropertyList
    init?(propertyListRepresentation:PropertyList?)
}

enum Happiness2: PropertyListReadable { // lets make our enum conform to "PropertyListReadable"
    case Sad
    case Happy(ofGrade: Int)
    
    init?(propertyListRepresentation: PropertyList?) {      // a new init with an optional dictionary as input
        guard let values = propertyListRepresentation else {return nil} // did we really get a dictionary?
        if let happiness = values["happiness"] as? Int {    // thats the main value of the enum
            switch happiness {
            case 0:                     // the first case mapped to 0
                self = .Sad
            case 1:                     // the second case mapped to 1
                if let grade = values["grade"] as? Int {    // and thats the associated value
                    self = .Happy(ofGrade: grade)
                } else {
                    return nil
                }
            default:                    // everything else is wrong
                return nil
            }
        } else {
            return nil
        }
    }
    
    func propertyListRepresentation() -> PropertyList {     // and the encoding part
        var happiness: Int              // lets map the cases of the enum to an Int
        var representation: PropertyList = [:]      // this will be the resulting dictionary
        switch self {
        case .Sad:                      // map first case to 0
            happiness = 0
        case .Happy(let grade):         // map second case to 1
            happiness = 1
            representation["grade"] = grade // and add the assciated value
        }
        representation["happiness"] = happiness // add the main value to the dictionary
        return representation
    }
}

let mood2 = Happiness2.Happy(ofGrade: 3)    // another instance of the enum
                                        // and the Data object of the enum
let data2 = NSKeyedArchiver.archivedData(withRootObject: mood2.propertyListRepresentation())

if let mood22 = Happiness2.init(propertyListRepresentation:
    NSKeyedUnarchiver.unarchiveObject(with: data2) as? PropertyList) {
    Swift.print("\(mood22)")            // the decoded enum
}

// ---------------------------------------------------------------------

extension Happiness2 {
    class Coding: NSObject, NSCoding {  // add a nested class to the enum that conforms to NSCoding
        let happiness: Happiness2?      // remember the parent enum
        
        init(happiness: Happiness2) {
            self.happiness = happiness  // set the parent enum
            super.init()
        }
        
        required init?(coder aDecoder: NSCoder) {   // initialize from a Coder (e.g. NSKeyedUnarchiver)
            guard let propertyListRepresentation = aDecoder.decodeObject() as? PropertyList
                else {return nil}
                                        // decode the dictionary
            happiness = Happiness2(propertyListRepresentation: propertyListRepresentation)
            super.init()                // and create the enum from it
        }
        
        public func encode(with aCoder: NSCoder) {  // encode the enum
            guard let happiness = happiness else {
                return
            }
                                        // just encode the dictionary
            aCoder.encode(happiness.propertyListRepresentation())
        }
    }
}

// MARK: - Encodable protocol

protocol Encodable {
    var encoded: Decodable? { get }
}
protocol Decodable {
    var decoded: Encodable? { get }
}

extension Happiness2: Encodable {
    var encoded: Decodable? {           // the NSCoding compliant object
        return Happiness2.Coding(happiness: self)
    }
}
extension Happiness2.Coding: Decodable {
    var decoded: Encodable? {           // the original enum after decoding the dirctionary
        return self.happiness
    }
}

let data3 = NSKeyedArchiver.archivedData(withRootObject: mood2.encoded!)
if let mood23 = (NSKeyedUnarchiver.unarchiveObject(with: data3) as? Happiness2.Coding)?.decoded {
    Swift.print("\(mood23)")
}
